/*
 * Copyright (C) 2001 Igor Sysoev.
 */

#include "mod_accel.h"

/*
   Return DECLINED or response status
*/

int accel_read_backend_header(accel_rec *a, int *backend)
{
    int              i, rc;
    header_rec      *bh;
    backend_connect  bc;

    request_rec *r = a->r;

    bc.r = r;
    bc.b = a->backend;
    bc.url = a->url;
    bc.addr.s_addr = a->pass->the_same_host ?
                                  r->connection->local_addr.sin_addr.s_addr:
                                  a->pass->addr.s_addr;
    bc.timeout = a->dir_conf->connect_timeout;
    bc.so_rcvbuf = a->dir_conf->bk_rcv_sock_buff_size;
    bc.backend = backend;
    bc.backends = a->backends;
    bc.failed = ap_make_array(a->r->pool, 3, sizeof(backend_failed));
    bc.path = NULL;
    bc.port_str = NULL;

#ifdef MOD_ACCEL_API
    if (!ap_hook_call("mod_accel::open_backend", &a->backend, &bc)
        || a->backend == NULL)
#endif
    {
        a->backend = accel_connect_to_http_backend(&bc);
    }

    if (bc.failed->nelts) {
        backend_failed *failed = (backend_failed *)bc.failed->elts;
        for (i = 0; i < bc.failed->nelts; i++) {
            ap_table_merge(a->ab->log_notes, "accel_r",
                           ap_pstrcat(r->pool, a->method, " ",
                                      inet_ntoa(failed[i].addr),
                                      bc.port_str, bc.path, " HTTP/1.0", NULL));

            if (a->backend || i < bc.failed->nelts - 1) {
                if ((a->backend_state = failed[i].state)
                                      == HTTP_GATEWAY_TIME_OUT)
                    a->backend_time = bc.timeout;
                else
                    a->backend_time = 0;

                accel_set_notes(a);

                a->backend_time = -1;

            } else {
                if (failed[i].state == HTTP_GATEWAY_TIME_OUT) {
                    a->backend_time = bc.timeout;
                    a->backend_timedout = 1;

                } else {
                    a->backend_time = 0;
                }
            }
        }

        a->backend_state = -1;
    }

    if (a->backend == NULL)
        return DECLINED;

    a->backend_time = a->time = time(NULL);
    a->backend_fd = a->backend->sock;
    a->backend_ip = ap_pstrdup(a->r->pool, inet_ntoa(a->backend->addr));

    ap_table_merge(a->ab->log_notes, "accel_r",
                   ap_pstrcat(r->pool, a->method, " ",
                              a->backend_ip, a->backend->port_str,
                              a->backend->path, " HTTP/1.0", NULL));

    if (!accel_send_request_to_backend(a)
        || !accel_read_http_header(a, BACKEND))
    {
        a->backend_aborted = 1;
        return DECLINED;
    }

    a->header = BACKEND;
    bh = a->backend_header;
    a->backend_state = bh->status;

#ifdef MOD_ACCEL_DEBUG
    accel_log_header(r, "backend", bh->line, bh->headers_out);
#endif

    /* ACCEL_PASS, ACCEL_NTNC or ACCEL_AUTH */

    if (!is_cachable(a))
        return bh->status;

    /* ACCEL_INVL, ACCEL_PGNC or ACCEL_MISS */

    if (a->accel_state >= ACCEL_INVL && a->accel_state <= ACCEL_MISS) {
        /* backend response can't be 304 in INVL, PGNC or MISS state
           because we always get full response but client may ask IMS */

        if (bh->status == HTTP_OK) {
            int lm = ap_parseHTTPdate(ap_table_get(bh->headers_out,
                                                   "Last-Modified"));

            if ((rc = accel_check_ims(a, lm, bh)) != OK) {
                a->cache_body = 1;
                return rc;
            }

        } else {
            return bh->status;
        }
    }

    /* ACCEL_AGED, ACCEL_EXPR or ACCEL_RVUS */

    if (bh->status == HTTP_NOT_MODIFIED) {
        /* if backend response is 304 then it's revalidating of our cache
           because we don't send client's IMS is this case */

        /* overlay backend headers to cache headeres */
        array_header *ha = ap_table_elts(bh->headers_out);
        table_entry *h = (table_entry *)ha->elts;
        for (i = 0; i < ha->nelts; i++)
            ap_table_setn(a->cache_header->headers_out, h[i].key, h[i].val);

        a->header = CACHE;
        a->body = CACHE;

        a->backend_closed = 1;
        accel_close_backend(a);

        if ((rc = accel_check_ims(a, a->cache_lm, a->cache_header)) != OK)
            return rc;

        lseek(a->cache_fd, a->cache_size, SEEK_SET);

        return a->cache_header->status;

    } else {
        a->body = BACKEND;
        a->cache_flags = 0;
        a->cache_len = 0;

        return bh->status;
    }
}

void accel_clean_backend_state(accel_rec *a)
{
    if (is_cachable(a))
        a->expired = -1;

    a->backend_closed = 0;
    a->backend_aborted = 0;
    a->backend_timedout = 0;

    a->backend_time = -1;
    a->backend_cache_time = -1;

    a->backend_first_read = -1;
    a->backend_second_read = 0;
    a->backend_read = 0;
    a->backend_reads = 0;
}


/*
   Send request to backend using HTTP/1.0 - we don't want chunked response.
   It's not problem in our case because we don't use keep-alive connections
   with backend.

   "Host" header set to backend hostname and port, if any.
   "Connection" header set to "close" to avoid persistent
                                 connection with backend.

   "If-Modified-Since" header set to cached "Last-Modified" header
                              if cached response is revalidated.
   Client's "If-Modified-Since", "If-Unmodified-Since", "If-Match" and
            "If-None-Match" headers hasn't sent if response can be cachable.

   "If-Range", "Range" and "Request-Range" headers hasn't sent
                              if request can be cachable.

   "Pragma" header hasn't sent if request can be cacheable
                                 and AccelIgnoreNoCache is set.
   "Cookie" header hasn't sent if request can be cachable
                                 and header is not revalidated.
*/

int accel_send_request_to_backend(accel_rec *a)
{
    int           i;
    char         *ims, *via_port, *ver;
    const char   *fwd, *via, *host, *x_url_host, *pass_host = NULL;
    BUFF         *fb;
    array_header *ha;
    table_entry  *h;

    int cookie_sent = 0;
    request_rec *r = a->r;
    backend_rec *b = a->backend;
    accel_dir_conf *dcf = a->dir_conf;

    fb = a->backend_fb = ap_bcreate(r->pool, B_RDWR | B_SOCKET);
    ap_bpushfd(fb, a->backend_fd, a->backend_fd);

    /* prepare all strings allocated from pool before accel_set_timeout() */

    if (dcf->x_url || dcf->x_host || a->pass->preserve_host)
        host = ap_table_get(r->headers_in, "Host");

    if (a->pass->preserve_host) {
        if (host) {
            char *port = strchr(host, ':');
            pass_host = port ? ap_pstrndup(r->pool, host, port - host) : host;
        }

    } else {
        pass_host = b->host;
    }

    if (a->revalidate || dcf->revalidate_user)
        ims = ap_gm_timestr_822(r->pool, a->cache_lm);

    if (dcf->x_fwd_for) {
        if (fwd = ap_table_get(r->headers_in, "X-Forwarded-For"))
            fwd = ap_pstrcat(r->pool, fwd, ", ", NULL);
        else
            fwd = "";
    }

    if (dcf->via) {
        int p = ap_get_server_port(r);
        via_port = (p == DEFAULT_HTTP_PORT) ? "":
                                              ap_psprintf(r->pool, ":%d", p);
        ver = ap_psprintf(r->pool, "%d.%d ", HTTP_VERSION_MAJOR(r->proto_num),
                                             HTTP_VERSION_MINOR(r->proto_num));

        if (via = ap_table_get(r->headers_in, "Via"))
            via = ap_pstrcat(r->pool, via, ", ", NULL);
        else
            via = "";
    }

    if (dcf->x_url) {
        int port = ap_get_server_port(r);

        x_url_host = host ? host : ap_get_server_name(r);

        if (port != DEFAULT_HTTP_PORT && strchr(x_url_host, ':') == NULL)
            x_url_host = ap_psprintf(r->pool, "%s:%d", x_url_host, port);
    }

    /* send request header */
    accel_set_timeout("accel: send request header to backend", a);

    ap_bvputs(fb, a->method, " ", b->path, " HTTP/1.0" CRLF, NULL);
    if (pass_host)
        ap_bvputs(fb, "Host: ", pass_host, b->port_str, CRLF, NULL);

    /* don't use persistent connection with backend */
    ap_bvputs(fb, "Connection: close" CRLF, NULL);

    if (a->cookie && !dcf->pass_cookie && !dcf->revalidate_user) {
        ap_bvputs(fb, "Cookie: ", a->cookie, CRLF, NULL);
        cookie_sent = 1;
    }

    if (a->revalidate || dcf->revalidate_user)
        ap_bvputs(fb, "If-Modified-Since: ", ims, CRLF, NULL);

    if (dcf->x_fwd_for)
        ap_bvputs(fb, "X-Forwarded-For: ", fwd,
                  r->connection->remote_ip, CRLF, NULL);

    if (dcf->via)
        ap_bvputs(fb, "Via: ", via, ver, ap_get_server_name(r), via_port,
                      " (" MOD_ACCEL_VER ")" CRLF, NULL);

    if (dcf->x_real_ip)
        ap_bvputs(fb, "X-Real-IP: ", r->connection->remote_ip, CRLF, NULL);

    if (dcf->x_host && host)
        ap_bvputs(fb, "X-Host: ", host, CRLF, NULL);

    if (dcf->x_url)
        ap_bvputs(fb, "X-URL: ", x_url_host, r->unparsed_uri, CRLF, NULL);

    if (dcf->x_uri)
        ap_bvputs(fb, "X-URI: ", r->unparsed_uri, CRLF, NULL);

    ha = ap_table_elts(r->headers_in);
    h = (table_entry *)ha->elts;
    for (i = 0; i < ha->nelts; i++) {
        if (h[i].key == NULL || h[i].val == NULL
            || strncasecmp(h[i].key, "Host", 4) == 0
            || strncasecmp(h[i].key, "Connection", 10) == 0
            || strncasecmp(h[i].key, "Keep-Alive", 10) == 0

            || (dcf->x_fwd_for
                && strncasecmp(h[i].key, "X-Forwarded-For", 15) == 0)
            || (dcf->via
                && strncasecmp(h[i].key, "Via", 3) == 0)

            || (is_cachable(a)
                && (strncasecmp(h[i].key, "If-Modified-Since", 17) == 0
                    || strncasecmp(h[i].key, "If-Unmodified-Since", 19) == 0
                    || strncasecmp(h[i].key, "If-Match", 8) == 0
                    || strncasecmp(h[i].key, "If-None-Match", 13) == 0

                    || strncasecmp(h[i].key, "If-Range", 8) == 0
                    || strncasecmp(h[i].key, "Range", 5) == 0
                    || strncasecmp(h[i].key, "Request-Range", 13) == 0

                    || (dcf->ignore_no_cache
                        && (strncasecmp(h[i].key, "Pragma", 6) == 0
                            || strncasecmp(h[i].key, "Cache-Control", 13) == 0)
                        )
                    || ((cookie_sent 
                           || (!dcf->revalidate_user && !dcf->pass_cookie))
                        && strncasecmp(h[i].key, "Cookie", 6) == 0)
                   )
               )
           )
            continue;

#ifdef MOD_ACCEL_DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
                      "accel: request header: '%s: %s'", h[i].key, h[i].val);
#endif

	ap_bvputs(fb, h[i].key, ": ", h[i].val, CRLF, NULL);
    }

    ap_bputs(CRLF, fb);
    accel_kill_timeout(a);

    /* send request data, if any */
    if (a->request_buf) {
        int pos, rlen, wlen;

        for (pos = 0; pos < r->read_length; pos += rlen) {
            if (a->request_file) {
                if ((rlen = read(a->request_fd, a->request_buf,
                                 dcf->request_buff_size))
                         <= 0) {
                    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
                                  "accel: can't read request body "
                                  "from tempfile %s",
                                  a->request_file);
                    return FALSE;
                }
                a->request_seek += rlen;

            } else {
                rlen = r->read_length;
            }

            accel_set_timeout("accel: send request data to backend", a);
            wlen = ap_bwrite(fb, a->request_buf, rlen);
            accel_kill_timeout(a);

            if (wlen != rlen) {
                ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
                              "accel: can't write to backend %s "
                              "while processing %s",
                              a->backend_ip, a->url);
                return FALSE;
            }
        }
    }

    accel_set_timeout("accel: send request data to backend", a);
    ap_bflush(fb);
    accel_kill_timeout(a);

    return TRUE;
}

void accel_send_backend_range(accel_rec *a)
{
    long off, size, len = 0;

    int range = a->cache_body ? 0 : 1;
    request_rec *r = a->r;
    char *buf = accel_alloca(a->send_threshold);

    while ((!a->cache_body
                 && range
                 && !r->connection->aborted)
             || (!a->backend_closed
                 && !a->backend_aborted)) {

        if (r->connection->aborted && (!is_cachable(a) || a->expired)) {
            a->backend_aborted = 1;
            accel_close_backend(a);
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, r,
                          "accel: backend is prematurely closed");
            break;
        }

        if (a->backend_nonblocked && r->connection->aborted) {
            if (ap_bblock(a->backend_fb, 0) == -1)
                accel_block_error(r, 1, 1);
            a->backend_nonblocked = 0;
        }

        if (!a->backend_closed && !a->backend_aborted) {
            if (a->backend_nonblocked) {
                if (a->client_writable)
                    accel_read_and_check_writable(a, NULL, 0);
                else
                    accel_read_and_check_writable(a, a->client_fb, 0);
            } else {
                accel_backend_read(a);
            }
        }

        if (len == 0 && range) {
            range = ap_each_byterange(r, &off, &len);
        }

        if (range && off < a->incnt) {
            size = 0;

            if (len >= a->send_threshold) {
                if (a->incnt - off >= a->send_threshold)
                    size = a->send_threshold;

            } else {
                if (a->incnt - off >= len)
                    size = len;
            }

            if (size) {
                size = accel_read_from_buffer(a, buf, off, size);

                ap_soft_timeout("accel: send range to client", r);
                ap_rwrite(buf, size, r);
                ap_kill_timeout(r);

                off += size;
                len -= size;
            }
        }
    }

    accel_freea(buf);
}

void accel_send_backend_body(accel_rec *a)
{
    int          rsize, wsize, first, tail, pos;

    request_rec *r = a->r;
    char *buf = accel_alloca(a->send_threshold);
    pos = 0;
    first = 1;

    /* cycle while we can write to client or can read from backend */

    while ((!a->cache_body
                 && (a->incnt - a->outcnt > 0)
                 && !r->connection->aborted)
             || (!a->backend_closed
                 && !a->backend_aborted)) {

        if (r->connection->aborted && (!is_cachable(a) || a->expired)) {
            a->backend_aborted = 1;
            accel_close_backend(a);
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, r,
                          "accel: backend is premature closed");
            break;
        }

        if (a->backend_nonblocked && r->connection->aborted) {
            if (ap_bblock(a->backend_fb, 0) == -1)
                accel_block_error(r, 1, 1);
            a->backend_nonblocked = 0;
        }

        if (!a->backend_closed && !a->backend_aborted) {
            if (a->backend_nonblocked) {
                if (a->client_writable)
                    accel_read_and_check_writable(a, NULL, 0);
                else
                    accel_read_and_check_writable(a, a->client_fb, 0);
            } else {
                accel_backend_read(a);
            }
        }

#ifdef MOD_ACCEL_DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
                      "accel: in:out %d:%d:%d",
                      a->incnt, a->outcnt, a->client_writable);
#endif

        if (!a->cache_body
             && a->client_writable
             && (a->incnt - a->outcnt > 0)
             && !r->connection->aborted) {

            if (a->incnt - a->outcnt >= a->send_threshold) {
                rsize = accel_read_from_buffer(a, buf + pos, a->outcnt + pos,
                                               a->send_threshold - pos);
                wsize = rsize - accel_content_tail + pos;
                tail = TRUE;

            } else if (a->backend_closed || a->backend_aborted) {
                rsize = accel_read_from_buffer(a, buf + pos, a->outcnt + pos,
                                               a->send_threshold);
                wsize = rsize + pos;
                tail = FALSE;

            } else {
                continue;
            }

#ifdef MOD_ACCEL_DEBUG
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
                          "accel: writing %d:%d:%d bytes to client",
                          rsize, wsize, tail);
#endif

#ifdef MOD_ACCEL_API
            a->cnt->buf = buf;
            a->cnt->len = wsize;
            a->cnt->state = 0;

            if (first) {
                a->cnt->state |= CNT_FIRST;
                first = 0;
            }
            if (tail)
                a->cnt->state |= CNT_TAIL;

            ap_hook_call("mod_accel::content", r, a->cnt);
#endif
            ap_soft_timeout("accel: send body to client", r);
            ap_rwrite(buf, wsize, r);
            ap_kill_timeout(r);

            a->outcnt += wsize;

            if (tail) {
                pos = accel_content_tail;
#ifdef MOD_ACCEL_DEBUG
                ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
                              "accel: copy tail %d", pos);
#endif
                memcpy(buf, buf + wsize, pos);
            }
        }
    }

    accel_freea(buf);

#ifdef MOD_ACCEL_DEBUG
    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
                  "accel: total read %d/%d, write %d bytes",
                  a->backend_second_read, a->incnt, a->outcnt);
#endif
}

/*
   write() and writev() EAPI hooks
*/

int accel_write_hook(BUFF *fb, char *buf, int len)
{
    int             rv;
    accel_buff_rec *ab;

    if ((ab = ap_ctx_get(fb->ctx, "mod_accel_buff"))
         && ab->a && ab->a->backend_fd >= 0) {
#ifdef MOD_ACCEL_DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, ab->a->r,
                      "accel: write hook entered %d", len);
#endif
        accel_read_and_check_writable(ab->a, fb, 1);
#ifdef MOD_ACCEL_DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, ab->a->r,
                      "accel: write hook exited");
#endif
    }

#ifdef MOD_SSL
    if (!ap_hook_call("mod_accel::write", &rv, fb, buf, len))
#endif
        rv =  write(fb->fd, buf, len);

    return rv;
}

int accel_writev_hook(BUFF *fb, const struct iovec *iov, int iovcnt)
{
    int             rv;
    accel_buff_rec *ab;

    if ((ab = ap_ctx_get(fb->ctx, "mod_accel_buff"))
         && ab->a && ab->a->backend_fd >= 0) {
#ifdef MOD_ACCEL_DEBUG
        if (iovcnt >= 2)
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, ab->a->r,
                          "accel: writev hook entered %d:%d:%d",
                          iovcnt, iov[0].iov_len, iov[1].iov_len);
        else
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, ab->a->r,
                          "accel: writev hook entered %d:%d",
                          iovcnt, iov[0].iov_len);
#endif
        accel_read_and_check_writable(ab->a, fb, 1);
#ifdef MOD_ACCEL_DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, ab->a->r,
                      "accel: writev hook exited");
#endif
    }

#ifdef MOD_SSL
    if (!ap_hook_call("mod_accel::writev", &rv, fb, iov, iovcnt))
#endif
        rv = writev(fb->fd, iov, iovcnt);

    return rv;
}

/*
   Read from backend if backend's fd is ready for reading.

   If backend had closed or aborted connection then return TRUE because
   after that client's fd would be in blocking mode.

   If accel_read_and_check_writable() was called while writing to client
   then ap_select() can be interrupted by ap_*_timeout(). Then next
   ap_select() should return immediately with EBADF because fb->fd closed
   by ap_*_timeout().

   If accel_read_and_check_writable() was called while reading from backend
   and both sockets are not ready then ap_select() would timeouted
   after AccelTimeout (by default equals to Timeout) seconds and we will
   close backend.
*/

void accel_read_and_check_writable(accel_rec *a, BUFF *fb, int hook)
{
    int                 fb_fd, rv;
    fd_set              in, out;
    struct timeval      tv;

    /* fb->fd can be set to -1 by ap_*_timeout()
       and we'll get signal 10 or 11 in FD_ISSET() */
    fb_fd = fb ? fb->fd: 0;

    if (fb_fd == -1) {
        a->client_writable = 0;
        return;
    }

    do {
        do {
            FD_ZERO(&in);
            FD_SET(a->backend_fd, &in);

            tv.tv_sec = a->dir_conf->timeout;
            tv.tv_usec = 0;

            if (fb) {
                FD_ZERO(&out);
                FD_SET(fb_fd, &out);
                rv = ap_select((a->backend_fd >= fb_fd ?
                                                 a->backend_fd : fb_fd) + 1,
                                &in, &out, NULL, &tv);
            } else {
                rv = ap_select(a->backend_fd + 1, &in, NULL, NULL, &tv);
            }

        } while (rv < 0 && (errno == EINTR || errno == EAGAIN));

        if (rv < 0) {
            if (errno == EBADF && fb && fb->fd == -1) { /* client timeouted */
                a->client_writable = 0;
                return;
            }

            ap_log_rerror(APLOG_MARK, APLOG_ERR, a->r, "accel: select failed");
        }

        if (rv == 0) {
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, a->r,
                          "accel: backend %s timeouted while processing %s",
                          a->backend_ip, a->url);
            a->backend_timedout = 1;
            a->backend_aborted = 1;
            accel_close_backend(a);
            return;
        }

        if (FD_ISSET(a->backend_fd, &in))
            if (!accel_backend_read(a))
                return;

    } while (hook && !FD_ISSET(fb_fd, &out));

    if (fb) {
        if (FD_ISSET(fb_fd, &out))
            a->client_writable = 1;
        else
            a->client_writable = 0;
    }
}

/*
   Read response from backend.

   If error has occured or EOF recieved then close backend's fd,
   set client's fd to blocking mode, make it writable and return FALSE.

   Return TRUE otherwise.
*/

int accel_backend_read(accel_rec *a)
{
    int n;

    do {
/*
        if (a->backend_fd < 0)
            return FALSE;
*/
        if (!a->backend_nonblocked)
            accel_set_timeout("accel: reading response from backend", a);

        n = ap_bread(a->backend_fb, a->buf + a->inptr,
                     a->bk_rcv_buff_size - a->inptr);

        a->backend_reads++;

        if (!a->backend_nonblocked)
            accel_kill_timeout(a);

#ifdef MOD_ACCEL_DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, a->r,
                      "accel: reading %d bytes from backend", n);
#endif

        if (n == -1) {
            if (errno == EAGAIN)
                return TRUE;

            a->backend_aborted = 1;
            ap_log_rerror(APLOG_MARK, APLOG_ERR, a->r,
                          "accel: error reading from backend %s "
                          "while processing %s",
                          a->backend_ip, a->url);
        }

        if (n == 0) {
            a->backend_closed = 1;
#ifdef MOD_ACCEL_DEBUG
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, a->r,
                          "accel: backend close socket", n);
#endif
        }

        if (n <= 0) {
            if (a->temp_fd >= 0)
                accel_tempfile_write(a, a->buf, a->inptr,
                                (off_t) (a->header_end + a->incnt - a->inptr));

            accel_close_backend(a);
            return FALSE;
        }

        if (a->backend_second_read == 0)
            a->backend_second_read = n;
        else
            a->backend_read += n;

        a->inptr += n;
        a->incnt += n;

        if (a->inptr == a->bk_rcv_buff_size) {

            if (a->temp_fd < 0) {
                ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_WARNING, a->r,
                              "accel: request %s is buffered to tempfile",
                              a->url);

                a->cache_state_line = cache_state_line_mask;
                if (!accel_open_tempfile(a, ACCEL_STUB))
                    return FALSE;
            }

            if (!accel_tempfile_write(a, a->buf, a->bk_rcv_buff_size,
                     (off_t) (a->header_end + a->incnt - a->bk_rcv_buff_size)))
                return FALSE;

            a->inptr = 0;
        }

    } while (a->inptr == 0);

    return TRUE;
}

void accel_close_backend(accel_rec *a)
{
#ifdef MOD_ACCEL_DEBUG
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, a->r,
                          "accel: backend CLOSED %d", a->backend_fd);
#endif

    if (a->backend_fd == -1)
        return;

    ap_bclose(a->backend_fb);       /* it also closes backend's socket       */

    a->time = time(NULL);
    a->backend_time = a->time - a->backend_time;

#ifdef MOD_ACCEL_API
    ap_hook_call("mod_accel::close_backend", a->backend_fd);
#endif
    a->backend_fd = -1;

    if (ap_bblock(a->client_fb, 0) == -1) /* client's fd is in blocking mode */
        accel_block_error(a->r, 0, 1);
    a->client_writable = 1;

    accel_response_expired(a);
    accel_update_cache(a);          /* update cache file and clear busy lock */
    accel_clear_busy_lock(a);
}

/*
    Read len bytes from buffer starting from pos.
    If part or all data is not available in memory then read it
    from temporary file.
*/

int accel_read_from_buffer(accel_rec *a, char *buf, int pos, int len)
{
    len = (pos + len > a->incnt) ? (a->incnt - pos) : len;

    if (pos < (a->incnt - a->inptr)) {
        /* part of data is already in temporary file */

        int rlen;
        off_t seek = a->header_end + pos;
        int flen = (a->incnt - a->inptr) - pos;
        flen = flen > len ? len : flen;

        if (a->temp_seek != seek)
            lseek(a->temp_fd, seek, SEEK_SET);

        if ((rlen = read(a->temp_fd, buf, flen)) != flen) {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, a->r,
                          "accel: can't read from tempfile %s", a->tempfile);
            return rlen;
        }

        a->temp_seek = seek + flen;

        if (len > flen)
            /* copy part of data that remains in memory */
            memcpy(buf + flen, a->buf, len - flen);

    } else {
        /* all data is in memory */
        memcpy(buf, (a->buf + pos - (a->incnt - a->inptr)), len);
    }

    return len;
}
