/*
 * Copyright (C) 2001 Igor Sysoev.
 */

#include "mod_accel.h"

/* not thread-safe */
static accel_rec *volatile accel;

static char *accel_states[] =
{
    "-", "PASS", "NTNC", "AUTH", "RMVD",
    "INVL", "PGNC", "MISS", "AGED", "EXPR", "RVUS", "HIT"
};

static char *lock_states[] =
{
    "-", "U",  "C",  "UC",  "W",  "UW",  "CW",  "UCW",
    "N", "UN", "CN", "UCN", "WN", "UWN", "CWN", "UCWN"
};

static char *backend_cache_states[] =
{
    "-", "NNC", "XAE", "CTL", "EXP", "MVD", "LMF", "ADE"
};

static char *content_flags[] =
{
    "-", "R", "Q", "RQ"
};

/*
    Temporary file operations
*/

char *accel_create_tempfile(accel_rec *a, char **tempfile, int *temp_fd)
{
    int   fd;
    pool *p = a->r->pool;
    char *name = ap_pstrcat(p, a->server_conf->cache_root, "/tmpXXXXXX", NULL);

    if (mktemp(name) == NULL) {
        return ap_psprintf(p, "accel: can't make tempfile name %s", name);
    }

    if ((fd = ap_popenf(p, name,
                        O_RDWR | O_CREAT | O_EXCL | O_BINARY, 0640)) == -1) {
        return ap_psprintf(p, "accel: can't create tempfile %s", name);
    }

    *tempfile = name;
    *temp_fd = fd;
    return NULL;
}

int accel_open_tempfile(accel_rec *a, int stub)
{
    int           i, rv;
    char         *header, *rc;
    struct pool  *pool;
    array_header *ha;
    table_entry  *h;

    request_rec *r = a->r;

    if ((rc = accel_create_tempfile(a, &a->tempfile, &a->temp_fd)) != NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, r, rc);
        a->cache_error = 1;
        a->backend_aborted = 1;
        accel_close_backend(a);
        return FALSE;
    }

    pool = ap_make_sub_pool(r->pool);
    header = ap_pstrcat(pool, a->cache_state_line, "\n",
                              "URL: ", a->key, "\n",
                              a->backend_header->line, "\n", NULL);

    ha = ap_table_elts(a->backend_header->headers_out);
    h = (table_entry *)ha->elts;
    for (i = 0; i < ha->nelts; i++) {
        if (h[i].key == NULL || h[i].val == NULL
            || strncasecmp(h[i].key, "Connection", 10) == 0
            || strncasecmp(h[i].key, "Set-Cookie", 10) == 0
           )
            continue;

	header = ap_pstrcat(pool, header, h[i].key, ": ", h[i].val, "\n", NULL);
    }

    header = ap_pstrcat(pool, header, "\n", NULL);
    a->header_end = strlen(header);

    /* some hack */
    if (!stub) {
        char *len;
        a->cache_len = a->header_end + a->incnt;
        len = ap_psprintf(r->pool, "%08X", a->cache_len);
        memcpy(&header[CACHE_LEN_POS], len, 8);
    }

    rv = accel_tempfile_write(a, header, a->header_end, (off_t) 0);
    ap_destroy_pool(pool);

    return rv;
}


int accel_tempfile_write(accel_rec *a, char *buf, int size, off_t seek)
{
    if (a->temp_seek != seek)
        lseek(a->temp_fd, seek, SEEK_SET);

    if (write(a->temp_fd, buf, size) != size) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, a->r,
                      "accel: can't write to tempfile %s", a->tempfile);
        a->cache_error = 1;
        a->backend_aborted = 1;
        accel_close_backend(a);
        return FALSE;
    }

    a->temp_seek = seek + size;

    return TRUE;
}

/*
   Timeout operations
*/

static void timeout(int sig);

void accel_set_timeout(char *timeout_name, accel_rec *a)
{
    accel = a;

    a->timeout_name = timeout_name;
    ap_set_callback_and_alarm(timeout, a->dir_conf->timeout);
}

void accel_kill_timeout(accel_rec *a)
{
    a->timeout_name = NULL;
    ap_set_callback_and_alarm(NULL, 0);
}

static void timeout(int sig)
{
     accel_rec *a = accel;

     a->backend_aborted = 1;
     a->backend_timedout = 1;

     if (a->backend_fb)
         a->backend_fb->flags |= B_EOUT;    /* ap_b*() */

     accel_close_backend(a);

     ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, a->r,
                  "[backend %s] %s timed out while processing %s",
                  a->backend_ip,
                  a->timeout_name ? a->timeout_name : "request",
                  a->url);
}


void accel_set_notes(accel_rec *a)
{
    char *accel_state, *expired_time, *lock_time, *lock_state,
         *backend_state, *backend_time,
         *backend_reads, *backend_first_read, *backend_read,
         *backend_cache, *backend_cache_time, *flags;

    int flag_mask = 0;
    pool *p = a->client_fb->pool;
    table *notes = a->ab->log_notes;

    accel_state = accel_states[a->accel_state];
    ap_table_merge(notes, "accel_st", accel_state);

    if (a->accel_state == ACCEL_RVUS || a->accel_state == ACCEL_EXPR)
        expired_time = a->expired_time < 0 ? "-":
                       ap_psprintf(p, "%d", a->expired_time);
    else
        expired_time = a->expired_time <= 0 ? "-":
                       ap_psprintf(p, "%d", a->expired_time);
    ap_table_merge(notes, "accel_et", expired_time);

    lock_time = a->lock_time == UNSET ? "-":
                                        ap_psprintf(p, "%d", a->lock_time);
    ap_table_merge(notes, "accel_lt", lock_time);

    lock_state = lock_states[a->lock_state];
    ap_table_merge(notes, "accel_ls", lock_state);

    backend_state = a->backend_state ? ap_psprintf(p, "%d", a->backend_state):
                          a->accel_state == ACCEL_HIT ? "HIT" : "-";
    ap_table_merge(notes, "accel_bs", backend_state);

    backend_time = a->backend_time < 0 ? "-" :
                   ap_psprintf(p, "%d", a->backend_time);
    ap_table_merge(notes, "accel_bt", backend_time);

    backend_first_read = a->backend_first_read < 0 ? "-" :
                         ap_psprintf(p, "%d", a->backend_first_read
                                              + a->backend_second_read);
    ap_table_merge(notes, "accel_bfr", backend_first_read);

    backend_read = a->backend_first_read < 0 ? "-" :
                 ap_psprintf(p, "%d", a->backend_first_read
                             + a->backend_second_read + a->backend_read);
    ap_table_merge(notes, "accel_br", backend_read);

    backend_reads = a->backend_reads <= 0 ? "-" :
                    ap_psprintf(p, "%d", a->backend_reads);
    ap_table_merge(notes, "accel_bnr", backend_reads);

    backend_cache = backend_cache_states[a->backend_cache];
    ap_table_merge(notes, "accel_bc", backend_cache);

    backend_cache_time = a->backend_cache_time == UNSET ? "-":
                         a->backend_cache_time == ACCEL_NO_CACHE ? "N":
                         ap_psprintf(p, "%d", a->backend_cache_time);
    ap_table_merge(notes, "accel_bct", backend_cache_time);

#ifdef MOD_ACCEL_API
    if ((a->cache_flags & (CNT_RANDBAN_CHECKED | CNT_RANDBAN_FOUND))
                       == (CNT_RANDBAN_CHECKED | CNT_RANDBAN_FOUND))
        flag_mask |= 1;

    if ((a->cache_flags & (CNT_QUOTED_CHECKED | CNT_QUOTED_FOUND))
                       == (CNT_QUOTED_CHECKED | CNT_QUOTED_FOUND))
        flag_mask |= 2;
#endif

    flags = content_flags[flag_mask];
    ap_table_merge(notes, "accel_fl", flags);

    ap_table_merge(notes, "accel", ap_psprintf(p,
                   "%s/%s/%s/%s %s/%s/%s %s %s/%s/%s %s",
                   accel_state, expired_time, lock_time, lock_state,
                   backend_state, backend_cache, backend_cache_time,
                   backend_time,
                   backend_reads, backend_first_read, backend_read, flags));
}

const char *accel_log(request_rec *r, char *note)
{
    accel_buff_rec *ab;

    if (ab = ap_ctx_get(r->connection->client->ctx, "mod_accel_buff"))
        return ap_table_get(ab->log_notes, note);

    return NULL;
}


int accel_max_age(pool *p, const char *string)
{
    char *token;
    const char *ptr = string;

    if (!string)
        return -1;

    while (*ptr && (token = ap_get_token(p, &ptr, 0))) {
        if (*ptr == ',' || *ptr == ';')         /* ap_get_token() bug */
            ptr++;

        /* BUG: look "max-age=x", but may be "max-age = " */
        if (strncmp(token, "max-age=", 8) == 0)
            return atoi(&token[8]);
    }

    return -1;
}


int accel_gets(char *buf, int size, accel_rec *a, int source)
{
    for ( ;; ) {
        int   len;
        char *p;

        for (p = a->temp_last; p < a->temp_end; p++) {
            int slen;

            if (*p == LF) {
                if ((slen = p - a->temp_last) > size)
                    return ACCEL_GETS_TOO_LONG;

                memcpy(buf, a->temp_last, slen);
                a->temp_last = p + 1;

                if (buf[slen - 1] == CR)
                    --slen;
                buf[slen] = '\0';

                return slen;
            }
        }

        if (source == BACKEND) {
            int temp_size = a->temp_size > a->dir_conf->bk_rcv_buff_size ?
                            a->dir_conf->bk_rcv_buff_size : a->temp_size;

            if (a->temp_buf < a->temp_last) {
                memcpy(a->temp_buf, a->temp_last, a->temp_end - a->temp_last);
                a->temp_end = a->temp_buf + (a->temp_end - a->temp_last);
                a->temp_last = a->temp_buf;
            }

            if (a->temp_end - a->temp_buf >= temp_size)
                return ACCEL_GETS_TOO_LONG;

            accel_set_timeout("accel: receive headers from backend", a);
            len = ap_bread(a->backend_fb, a->temp_end,
                           temp_size - (a->temp_end - a->temp_buf));
            accel_kill_timeout(a);

#ifdef MOD_ACCEL_DEBUG
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, a->r,
                          "accel: reading %d bytes from backend", len);
#endif   

        } else {
            if (a->temp_buf != a->temp_end)
                return ACCEL_GETS_TOO_LONG;

            len = read(a->cache_fd, a->temp_buf, a->temp_size);
            a->cache_size = a->cache_seek = len;
        }

        if (len < 0)
            return ACCEL_GETS_ERROR;

        if (len == 0)
            return ACCEL_GETS_EOF;

        if (source == BACKEND) {
            a->backend_reads++;
            if (a->backend_first_read == -1)
                a->backend_first_read++;
            a->backend_first_read += len;
        }

        a->temp_end += len;
    }
}

int accel_lookc(char *next, accel_rec *a)
{
    if (a->temp_last == a->temp_end) {
        int len;
        int temp_size = a->temp_size > a->dir_conf->bk_rcv_buff_size ?
                        a->dir_conf->bk_rcv_buff_size : a->temp_size;

        if (a->temp_buf < a->temp_last) {
            memcpy(a->temp_buf, a->temp_last, a->temp_end - a->temp_last);
            a->temp_end = a->temp_buf + (a->temp_end - a->temp_last);
            a->temp_last = a->temp_buf;
        }

        if (a->temp_end - a->temp_buf >= temp_size)
            return ACCEL_GETS_TOO_LONG;

        accel_set_timeout("accel: receive headers from backend", a);
        len = ap_bread(a->backend_fb, a->temp_end,
                       temp_size - (a->temp_end - a->temp_buf));
        accel_kill_timeout(a);

#ifdef MOD_ACCEL_DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, a->r,
                      "accel: reading %d bytes from backend", len);
#endif   

        if (len < 0)
            return ACCEL_GETS_ERROR;

        if (len == 0)
            return ACCEL_GETS_EOF;

        a->backend_reads++;
        a->backend_first_read += len;

        a->temp_end += len;
    }

    *next = *(a->temp_last);
    return 1;
}

void accel_gets_error(accel_rec *a, int source, int status,
                      int error, int save_errno)
{
    char        *src, *url;
    request_rec *r = a->r;

    if (source == BACKEND) {
        src = ap_pstrcat(r->pool, "backend ", a->backend_ip, NULL);
        url = ap_pstrcat(r->pool, " while processing ", a->url, NULL);

    } else {
        src = ap_pstrcat(r->pool, "cache ", a->cachefile, NULL);
        url = "";
    }

    if (error == ACCEL_GETS_EOF && !status)
        error = ACCEL_GETS_ERROR;

    switch (error) {

    case ACCEL_GETS_TOO_LONG:
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
                      "accel: header line from %s is bigger than %d%s",
                      src, a->temp_size, url);
        break;

    case ACCEL_GETS_ERROR:
        if (!a->backend_timedout) {
            errno = save_errno;
            ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
                          "accel: error reading header from %s%s", src, url);
        }
        break;

    case ACCEL_GETS_EOF:
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
                      "accel: document from %s contains no data%s", src, url);

    }
}

void accel_block_error(request_rec *r, int socket, int block)
{
    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
                  "accel: can not set %sblocking mode for %s's socket",
                  block ? "" : "non-", socket ? "backend" : "client");
}


void *accel_alloc(array_header *allocated, int size)
{
    void *ptr;

    if (ptr = malloc(size)) {
        char **new = ap_push_array(allocated);
        *(new) = ptr;
        return ptr;

    } else {
        return NULL;
    }
}

void accel_free(array_header *allocated, void *ptr)
{
    int i;
    void **block = (void **)allocated->elts;

    for (i = 0; i < allocated->nelts; i++) {
        if (block[i] == ptr)
            block[i] = NULL;
    }

    free(ptr);
}

void accel_cleanup_allocated(array_header *allocated)
{
    int i;
    void **block = (void **)allocated->elts;

    for (i = 0; i < allocated->nelts; i++) {
        if (block[i]) {
            free(block[i]);
            block[i] = NULL;
        }
    }
}


void *accel_set_time_string(pool *p, const char *name,
                            const char *string, int *value)
{
    int number, factor;
    int error = 0;
    int one_number = 1;

    *value = 0;

    while(*string) {
        char *token = ap_getword_conf(p, &string);

        if((number = atoi(token)) < 0)
           return ap_psprintf(p, "%s: bad number '%s'", name, token);

        token = ap_getword_conf(p, &string);

        if (!token || !*token) {
            if (one_number) {
                *value = number;
                return NULL;
            } else {
                return ap_psprintf(p, "%s: bad time name '%s'", name, token);
            }
        }

        one_number = 0;

        factor = 0;
        switch(token[0]) {
        case 'Y': case 'y':
            factor = 365 * 24 * 3600;
            break;
        case 'W': case 'w':
            factor = 7 * 24 * 3600;
            break;
        case 'D': case 'd':
            factor = 24 * 3600;
            break;
        case 'H': case 'h':
            factor = 3600;
            break;
        case 'S': case 's':
            factor = 1;
            break;
        case 'M': case 'm':
            switch(token[1]) {
            case 'O': case 'o':
                factor = 30 * 24 * 3600;
                break;
            case 'I': case 'i':
                factor = 60;
                break;
            default:
                error = 1;
            }
            break;
        default:
            error = 1;
        }

        if (error)
           return ap_psprintf(p, "%s: bad time name '%s'", name, token);
        else
           *value += factor * number;
    }

    return NULL;
}

int accel_sum(accel_sum_e part, char *url, const char *tag)
{
    char *start, *end;
    int sum = 0;

    if (part == ACCEL_SUM_NONE)
        return 0;

    if (part == ACCEL_SUM_TAG) {
        if (tag == NULL)
            return 0;

        while (*tag != '\0') {
            if (sum & 1)
                sum |= 0x10000;

            sum = ((sum >> 1) + *tag++) & 0xffff;
        }

        return sum;
    }

    if (!(start = strstr(url, "://")))
        return 0;

    start += 3;

    if (part == ACCEL_SUM_HOST) {
        if (!((end = strchr(start, ':')) || (end = strchr(start, '/'))))
            return 0;

    } else if (part == ACCEL_SUM_PORT) {
        if (!(end = strchr(start, '/')))
            return 0;

    } else { /* part == ACCEL_SUM_LOCATION */
        end = strchr(start, '\0');
    }

    while (start < end) {
        if (sum & 1)
            sum |= 0x10000;

        sum = ((sum >> 1) + *start++) & 0xffff;
    }

    return sum;
}

char *accel_rewrite_url(accel_rec *a, char *url)
{
    int             i, url_len;
    accel_pass     *pass;
    accel_reverse  *reverse;

    url_len = strlen(url);

    if (a->external) {
        reverse = (accel_reverse *)a->server_conf->reverse->elts;

        for (i = 0; i < a->server_conf->reverse->nelts; i++) {
            if (url_len >= reverse[i].url_len
                && strncmp(reverse[i].url, url, reverse[i].url_len) == 0)

                return ap_construct_url(a->r->pool, ap_pstrcat(a->r->pool,
                                                    reverse[i].loc,
                                                    &url[reverse[i].url_len],
                                                    NULL),
                                        a->r);
        }

    } else {
        pass = (accel_pass *)a->server_conf->pass->elts;

        for (i = 0; i < a->server_conf->pass->nelts; i++) {
            if (url_len >= pass[i].url_len
                && strncmp(pass[i].url, url, pass[i].url_len) == 0)

                return ap_construct_url(a->r->pool, ap_pstrcat(a->r->pool,
                                                    pass[i].loc,
                                                    &url[pass[i].url_len],
                                                    NULL),
                                        a->r);
        }
    }

    return NULL;
}

char *accel_rewrite_port(accel_rec *a, char *url, int our_port)
{
    int   p;
    char *prefix, *port, *pos, *host_end, *rest;

    if (!(prefix = strstr(url, "://")))
        return NULL;

    prefix += 3;

    port = strchr(prefix, ':');
    if (!port && a->pass->port != DEFAULT_HTTP_PORT)
        return NULL;

    if (port && (p = atoi(port + 1)) && p != a->pass->port)
        return NULL;

    host_end = port;
    pos = port ? port : prefix;

    if (!(rest = strchr(pos, '/')))
        rest = strchr(pos, '\0');

    if (!host_end)
        host_end = rest;

    return ap_pstrcat(a->r->pool, ap_pstrndup(a->r->pool, url, host_end - url),
                                  (our_port == DEFAULT_HTTP_PORT) ? "":
                                      ap_psprintf(a->r->pool, ":%d", our_port),
                                  rest, NULL);
}

#ifdef MOD_ACCEL_DEBUG
void accel_log_header(request_rec *r, char *name, char *line, table *headers)
{
    int           i;
    array_header *ha;
    table_entry  *h;

    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
                  "accel: %s status: %s", name, line);

    ha = ap_table_elts(headers);
    h = (table_entry *)ha->elts;
    for (i = 0; i < ha->nelts; i++) {
         ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
                           "accel: %s header: '%s: %s'",
                           name, h[i].key, h[i].val);
        }
    }
#endif

