/*
 * Copyright (C) 2001 Igor Sysoev.
 */

#include "mod_accel.h"

static void accel_cleanup(void *data);

int accel_content_tail = 0;

#ifdef EAPI_MM
/* not thread-safe */
int         accel_slot_num = -1;
pool       *shared_pool;
accel_slot *accel_state;
#endif

static void accel_add_module(module *m)
{
    if (m != &accel_module)
        return;

#ifdef MOD_ACCEL_API
    ap_hook_configure("mod_accel::open_backend",
                      AP_HOOK_SIG2(ptr,ptr), AP_HOOK_DECLINE(NULL));
    ap_hook_configure("mod_accel::close_backend",
                      AP_HOOK_SIG2(void,int), AP_HOOK_DECLINE(NULL));

    ap_hook_configure("mod_accel::ims",
                      AP_HOOK_SIG3(int,ptr,ptr), AP_HOOK_DECLINE(OK));
    ap_hook_configure("mod_accel::range",
                      AP_HOOK_SIG3(int,ptr,ptr), AP_HOOK_DECLINE(OK));
    ap_hook_configure("mod_accel::headers",
                      AP_HOOK_SIG3(void,ptr,ptr), AP_HOOK_ALL);
    ap_hook_configure("mod_accel::content",
                      AP_HOOK_SIG3(void,ptr,ptr), AP_HOOK_ALL);

    ap_hook_configure("mod_accel::cont_tail",
                      AP_HOOK_SIG3(void,ptr,ptr), AP_HOOK_ALL);
#endif

    ap_hook_register("ap::buff::write", accel_write_hook, AP_HOOK_NOCTX);
#ifndef NO_WRITEV
    ap_hook_register("ap::buff::writev", accel_writev_hook, AP_HOOK_NOCTX);
#endif

#ifdef MOD_SSL
    ap_hook_configure("mod_accel::write",
                      AP_HOOK_SIG4(int,ptr,ptr,int), AP_HOOK_TOPMOST);
#ifndef NO_WRITEV
    ap_hook_configure("mod_accel::writev",
                      AP_HOOK_SIG4(int,ptr,ptr,int), AP_HOOK_TOPMOST);
#endif
#endif

    ap_hook_register("ap::mod_log_config::log_x", accel_log, AP_HOOK_NOCTX);
}

static void accel_remove_module(module *m)
{
    if (m != &accel_module)
        return;

    ap_hook_unregister("ap::buff::write", accel_write_hook);
#ifndef NO_WRITEV
    ap_hook_unregister("ap::buff::writev", accel_writev_hook);
#endif

    ap_hook_unregister("ap::mod_log_config::log_x", accel_log);
}

static void accel_init(server_rec *s, pool *p)
{
    accel_server_conf *conf = (accel_server_conf *)
                       ap_get_module_config(s->module_config, &accel_module);

    ap_add_version_component(MOD_ACCEL_VER);

#ifdef MOD_ACCEL_DEBUG
    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, s,
                 "accel: pid: %d ppid: %d", getpid(), getppid());
#endif

    if (conf->cache_root) {
        accel_gc_conf *agc = ap_palloc(p, sizeof(accel_gc_conf));

        agc->server = s;
        agc->pool = p;
        agc->ppid = getppid();
        agc->conf = conf;

        ap_start_other_child(s, p, accel_gc_init, (void *)agc);

    } else {
        ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ALERT, s,
                     "accel: AccelCacheRoot must be specified");
        exit(1);
    }

#ifdef EAPI_MM
    /* We should not make shared pool from pconf pool - p
       because in pconf pool there is one more pool in childs
       before our shared pool. This causes segfault in the parent
       while graceful restart at least when all child slots are busy.
       Instead we create intermidient pool and make shared pool from it.
    */
    shared_pool = ap_make_shared_sub_pool(ap_make_sub_pool(p));
    accel_state = ap_pcalloc(shared_pool,
                             ap_daemons_limit * sizeof(accel_slot));
#endif

#ifdef MOD_ACCEL_API
    ap_hook_call("mod_accel::cont_tail", s, &accel_content_tail);
#ifdef MOD_ACCEL_DEBUG
    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, s,
                 "accel: content tail length: %d", accel_content_tail);
#endif
#endif
}

/*
   Post read request handler clears mod_accel_buff context
   if connection is kept alive after last request.
   We need to check following conditions:
       subrequest (SSI with mod_rewrite) - r->main is not NULL;
       internal redirect (ErrorDocument) - r->main is NULL
                                           but mod_accel CTX exists.
*/

static int accel_postread(request_rec *r)
{
    accel_rec      *a;
    accel_buff_rec *ab;

    if (!r->main
        && !(a = ap_ctx_get(r->ctx, "mod_accel"))
        && (ab = ap_ctx_get(r->connection->client->ctx, "mod_accel_buff")))
    {
        ab->a = NULL;
        ap_clear_table(ab->log_notes);
    }

    return DECLINED;
}

/*
   Translatation handler looks if we need to proxy uri
   but it doesn't translate r->uri to a->url now - it waits
   response phase because mod_charset recodes r->uri in fixup phase.
*/

static int accel_translate(request_rec *r)
{
    int                i, j;
    accel_rec         *a;
    accel_pass        *pass;
    accel_no_pass     *no_pass;

    accel_server_conf *conf = (accel_server_conf *)
                ap_get_module_config(r->server->module_config, &accel_module);

    /* AccelPass */
    int uri_len = strlen(r->uri);
    pass = (accel_pass *) conf->pass->elts;

    for (i = 0; i < conf->pass->nelts; i++) {
        if (uri_len == pass[i].loc_len - 1
            && pass[i].loc[pass[i].loc_len - 1] == '/'
            && strncmp(r->uri, pass[i].loc, pass[i].loc_len - 1) == 0)
        {
            ap_table_setn(r->headers_out, "Location",
                          ap_construct_url(r->pool,
                                           ap_pstrcat(r->pool,
                                                      r->uri, "/", NULL),
                                           r));
            return HTTP_MOVED_PERMANENTLY;
        }

        if (uri_len < pass[i].loc_len
            || strncmp(r->uri, pass[i].loc, pass[i].loc_len) != 0)
        {
            continue;
        }

        /* AccelNoPass */
        no_pass = (accel_no_pass *) conf->no_pass->elts;
        for (j = 0; j < conf->no_pass->nelts; j++) {
            if (no_pass[j].regexp) {
                if (ap_regexec(no_pass[j].regexp, r->uri, 0, NULL, 0) == 0)
                    return DECLINED;

            } else {
                if (uri_len < no_pass[j].len)
                    continue;

                if (strncmp(r->uri, no_pass[j].loc, no_pass[j].len) == 0)
                    return DECLINED;
            }
        }

        a = accel_init_ctx(r);

        r->proxyreq = ACCEL;
        r->handler = "accel-handler";

        a->pass = &pass[i];
        /* we don't make full URL at this stage because mod_charset can
           recode r->args and perhaps r->uri and r->path_info */
        a->url = pass[i].url;
        a->rest = r->uri + pass[i].loc_len;

        return OK;
    }

    return DECLINED;
}

/*
   "accel-handler" set in translation phase by mod_accel or by mod_rewrite.
*/

static int accel_handler(request_rec *r)
{
    int             rc, cached, backend, range = 0;
    char            temp_buf[TEMP_SIZE];
    accel_rec      *a;
    accel_dir_conf *dcf;
    int (*rewrite) (accel_rec *a);

    if (a = ap_ctx_get(r->ctx, "mod_accel")) {
        a->url = accel_make_url(r, ap_pstrcat(r->pool,
                                        a->url,
                                        ap_os_escape_path(r->pool, a->rest, 1),
                                        NULL));
#ifdef EAPI_MM
        if (a->pass->the_same_host && a->pass->sum_part == ACCEL_SUM_NONE)
            a->pass->sum = accel_sum(ACCEL_SUM_TAG, NULL,
                                     r->connection->local_ip);
        else if (a->pass->preserve_host && a->pass->sum_part == ACCEL_SUM_NONE)
            a->pass->sum = accel_sum(ACCEL_SUM_TAG, NULL,
                                     ap_table_get(r->headers_in, "Host"));
#endif

    } else {
#ifdef EAPI_MM
        const char *max, *tag;
#endif
        /* if it's "proxy:" request then we need to make new accel_rec */
        if (strncmp(r->filename, "proxy:", 6) != 0)
            return DECLINED;

        a = accel_init_ctx(r);
        a->external = 1;
        a->url = accel_make_url(r, r->filename + 6);
        a->pass = ap_pcalloc(r->pool, sizeof(accel_pass));

#ifdef EAPI_MM
        if (max = ap_table_get(r->notes, "rewrite_mc"))
            a->pass->max_conn = atoi(max);

        if (max = ap_table_get(r->notes, "rewrite_mw"))
            a->pass->max_wait = atoi(max);

        if (tag = ap_table_get(r->notes, "rewrite_mp")) {
            if (tag[0] == 'H')
                a->pass->sum_part = ACCEL_SUM_HOST;

            else if (tag[0] == 'P')
                a->pass->sum_part = ACCEL_SUM_PORT;

            else if (tag[0] == 'T') {
                a->pass->sum_part = ACCEL_SUM_TAG;
                tag++;
            }

        } else {
            a->pass->sum_part = ACCEL_SUM_HOST;
        }

        if (a->pass->max_conn && !a->pass->max_wait)
            a->pass->max_wait = a->pass->max_conn;

        a->pass->sum = accel_sum(a->pass->sum_part, a->url, tag);
#endif
    }

    a->request_buf = (char *) ap_table_get(r->notes, "accel_request_body");

    a->dir_conf = dcf = (accel_dir_conf *)
                ap_get_module_config(r->per_dir_config, &accel_module);

    accel_init_dir_config(a, dcf);

#ifdef MOD_ACCEL_DEBUG0
    accel_log_header(r, "client",
                     ap_psprintf(r->pool, "%s %s", r->method, a->url),
                     r->headers_in);
#endif

#ifdef MOD_ACCEL_API
    a->cnt = ap_pcalloc(r->pool, sizeof(accel_cnt));
#endif
    a->time = time(NULL);

    if (dcf->invl_len > 0) {
        int len = strlen(a->url) - dcf->invl_len;

        if (len >= 0 && strcmp(a->url + len, dcf->invalidate) == 0) {
            *(a->url + len) = '\0';
            a->invalidate = 1;
        }
    }

    /* response is cached ? */
    if (cached = accel_check_cachable(a)) {
        a->cache_buf = accel_alloca(dcf->cache_send_size + accel_content_tail);
        a->temp_buf = a->temp_last = a->temp_end = a->cache_buf;
        a->temp_size = dcf->cache_send_size;

        if (cached = accel_check_cache(a))
            a->expired = 0;
    }

    if (!is_cachable(a))
        a->expired = 1;

    if ((dcf->unlink_no_cached || a->invalidate) && !a->md5)
        accel_make_cachefile_name(a, 0);

    if (a->max_connected)
        a->lock_state |= ACCEL_MAX_CONN;
    if (a->max_waiting)
        a->lock_state |= ACCEL_MAX_WAIT;

    /* check IMS-like request */
    if (cached && a->cache_header->status == HTTP_OK) {

        if ((rc = accel_check_ims(a, a->cache_lm, a->cache_header)) != OK) {

            if (rc == HTTP_NOT_MODIFIED) {
#ifdef MOD_ACCEL_API
                ap_hook_call("mod_accel::headers", r, a->cache_header);
#endif
                accel_set_http_header(a, rc);
            }

            accel_update_cachefile(a);
            accel_set_notes(a);

            if (a->cache_buf)
                accel_freea(a->cache_buf);

            return rc;
        }

        rc = a->cache_header->status;
    }

    /* response is not cached or must be revalidate */
    if (!cached) {

        if (!r->main && !a->request_buf) {
            if ((rc = ap_setup_client_block(r, REQUEST_CHUNKED_ERROR)) != OK)
                return rc;

            if (ap_should_client_block(r)) {
                a->request_buf = accel_alloca(dcf->request_buff_size);

                if ((rc = accel_read_request_body(a)) != OK) {
                    accel_request_body_done(a);
                    return rc;
                }
            }
        }

        /* too many connections to backend */
        if ((a->pass->max_conn && a->max_connected_checked && a->max_connected)
            || (!a->busy_lock
                && accel_max_connected(a)))
        {
            a->lock_state |= ACCEL_MAX_CONN;
            if (a->max_waiting)
                a->lock_state |= ACCEL_MAX_WAIT;

            a->backend_state = HTTP_SERVICE_UNAVAILABLE;
            accel_set_notes(a);
            if (a->cache_buf)
                accel_freea(a->cache_buf);
            accel_request_body_done(a);

            return HTTP_SERVICE_UNAVAILABLE;
        }

        if (is_cachable(a) && r->header_only) {
            a->method = "GET";
            a->cache_body = 1;
        } else {
            a->method = r->method;
        }

        /* cycle through available backends */
        for (backend = 0; ; accel_clean_backend_state(a)) {
#ifdef MOD_ACCEL_DEBUG
           ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
                         "accel: backend: %d", backend);
#endif
            a->temp_buf = a->temp_last = a->temp_end = temp_buf;
            a->temp_size = TEMP_SIZE;

            if (a->request_seek) {
                lseek(a->request_fd, (off_t) 0, SEEK_SET);
                a->request_seek = 0;
            }

            rc = accel_read_backend_header(a, &backend);

            if ((rc == HTTP_NOT_MODIFIED || rc == HTTP_PRECONDITION_FAILED)
                && is_cachable(a) && !a->cache_body)
            {
                if (rc == HTTP_NOT_MODIFIED) {
#ifdef MOD_ACCEL_API
                    ap_hook_call("mod_accel::headers", r,
                                 a->header == BACKEND ? a->backend_header:
                                                        a->cache_header);
#endif
                    accel_set_http_header(a, rc);
                }

                a->backend_closed = 1;
                accel_close_backend(a);
                accel_set_notes(a);
                if (a->cache_buf)
                    accel_freea(a->cache_buf);

                return rc;
            }

            if (rc == DECLINED
                || (dcf->retry_5XX && ap_is_HTTP_SERVER_ERROR(rc))) {

                a->backend_state = a->backend_timedout ? HTTP_GATEWAY_TIME_OUT:
                                        rc == DECLINED ? HTTP_BAD_GATEWAY:
                                                         rc;
                if (backend < a->backends->nelts) {
                    accel_close_backend(a);
                    accel_set_notes(a);
                    continue;
                }

                if (a->backend_timedout || rc == DECLINED) {
                    a->backend_closed = 1;
                    accel_close_backend(a);
                    accel_set_notes(a);
                    if (a->cache_buf)
                        accel_freea(a->cache_buf);
                    accel_request_body_done(a);

                    return a->backend_state;
                }
            }

            break;
        }
    }

#ifdef MOD_ACCEL_API
    ap_hook_call("mod_accel::headers", r,
                (a->header == BACKEND) ? a->backend_header : a->cache_header);
#endif

    accel_request_body_done(a);

    accel_response_expired(a);

    /* set reading from backend and writing to client in non-blocking mode */
    if (a->body == BACKEND) {
        int rest = a->temp_end - a->temp_last;

        if (!a->backend_closed) {
            if (ap_bnonblock(a->backend_fb, 0) == -1)
                accel_block_error(r, 1, 0);
            a->backend_nonblocked = 1;
            if (ap_bnonblock(a->client_fb, 0) == -1)
                accel_block_error(r, 0, 0);
        }

        /* We can't alloca() a->buf because it probably can be used
            after accel_handler completed */

        a->bk_rcv_buff_size = dcf->bk_rcv_buff_size;
        a->buf = accel_alloc(a->allocated, a->bk_rcv_buff_size);

        if (rest > 0) {
            memcpy(a->buf, a->temp_last, rest);
            a->inptr = a->incnt = rest;
        }
    }

    /* check ranges in cachable responses */
    if (is_cachable(a)
        && rc == HTTP_OK
#ifdef MOD_ACCEL_API
        && (a->cache_flags & (CNT_RANGE_CHECKED | CNT_RANGE_DISABLED))
                          != (CNT_RANGE_CHECKED | CNT_RANGE_DISABLED)
#endif
    ) {
        /* r->clength needed for ap_set_byterange() */
        if (a->body == CACHE) {
            r->clength = a->cache_len - a->cache_header_size;

        } else {
            int len;
            const char *clen;

            if ((clen = ap_table_get(a->backend_header->headers_out,
                                     "Content-Length"))
                 && (len = atol(clen)) > 0
#ifdef MOD_ACCEL_API
                 && (a->cache_flags & (CNT_RANGE_CHECKED | CNT_RANGE_DISABLED))
                                    != CNT_RANGE_CHECKED
                 && (!ap_hook_call("mod_accel::range",
                                   &rc, a->r, a->backend_header))
#endif
               ) {
                r->clength = len;
            }
        }

        if(range = ap_set_byterange(r)) {
            rc = r->status;

            if (rc == HTTP_RANGE_NOT_SATISFIABLE && a->body == BACKEND)
                a->cache_body = 1;
        }
    }

    if ((rewrite = (void *)ap_table_get(r->notes, "accel_rewrite_response"))
        && (rc = (*rewrite)(a)) != DECLINED)
    {
        return rc;
    }

    /*
    We don't like hard timeouts in this handler because we need to save
    cached response even the connection to client is aborted so we use
    ap_soft_timeout() only.

    Although ap_send_http_header() calls ap_hard_timeout() the timeouts
    should be very rare because most headers are less than 4K and
    are buffered in Apache's BUFF. Besides even header is slightly bigger
    than 4K and a real write() is made there is kernel buffering and write()
    should not be blocked. But even it this case response should be
    cached if it would has recieved before ap_hard_timeout() will alarm.
    */

    if (is_cachable(a)) {
        if (rc != HTTP_PRECONDITION_FAILED)
            accel_set_http_header(a, rc);

        if (rc != HTTP_NOT_MODIFIED && rc != HTTP_PRECONDITION_FAILED)
            ap_send_http_header(r);

    } else {
        accel_set_http_header(a, rc);
        ap_send_http_header(r);
    }

#ifdef MOD_ACCEL_DEBUG0
    accel_log_header(r, "response", r->status_line, r->headers_out);
#endif

    if (a->body == CACHE && (r->header_only || a->cachemgr)) {
        accel_set_notes(a);
        if (a->cache_buf)
            accel_freea(a->cache_buf);

        if (rc == HTTP_NOT_MODIFIED)
            return HTTP_NOT_MODIFIED;
        else
            return OK;
    }

    if (a->body == BACKEND) {
        if (dcf->wait_before_body_read
            && (rc == HTTP_OK || rc == HTTP_PARTIAL_CONTENT))
        {
            usleep(dcf->wait_before_body_read * 1000);
        }

        a->send_threshold = dcf->send_size + accel_content_tail;

        if (range)
            accel_send_backend_range(a);
        else
            accel_send_backend_body(a);

    } else {
        if (range) {
            accel_send_cache_range(a);
        } else {
            a->send_threshold = dcf->cache_send_size + accel_content_tail;
            accel_send_cache_body(a);
        }
    }

    if (a->cache_buf)
        accel_freea(a->cache_buf);

    if (a->buf) {
        accel_free(a->allocated, a->buf);
        a->buf = NULL;
    }

#ifdef MOD_ACCEL_DEBUG
    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
                  "accel: flags: %d %d:%d",
                  a->cache_fd, a->cache_flags, a->cnt->out_flags);
#endif

    /* BUG: need to update last accessed time ? */

#ifdef MOD_ACCEL_API
    if (a->cache_fd >= 0 && a->cache_flags != a->cnt->out_flags) {
        a->cache_flags = a->cnt->out_flags;
        accel_update_cachefile(a);
    }
#endif

    if (a->cache_fd >= 0 && a->cache_fd != a->temp_fd)
         ap_pclosef(r->pool, a->cache_fd);

    a->cache_fd = -1;

    if (a->temp_fd >= 0) {
         ap_pclosef(r->pool, a->temp_fd);
         a->temp_fd = -1;
    }

    accel_set_notes(a);

    if (is_cachable(a))
        if (rc == HTTP_NOT_MODIFIED || rc == HTTP_PRECONDITION_FAILED)
            return rc;
        else
            return OK;
    else
        return OK;
}

accel_rec *accel_init_ctx(request_rec *r)
{
    accel_buff_rec *ab;

    accel_rec *a = ap_pcalloc(r->pool, sizeof(accel_rec));

    a->server_conf = (accel_server_conf *)
                ap_get_module_config(r->server->module_config, &accel_module);

    a->r = r;
    a->client_fb = r->connection->client;

    a->allocated = ap_make_array(r->pool, 3, sizeof(void *));
    a->backends = ap_make_array(r->pool, 5, sizeof(struct in_addr));

    a->request_fd = -1;
    a->backend_fd = -1;
    a->cache_fd = -1;
    a->temp_fd = -1;

    a->cachefile_existed = -1;
    a->expired = -1;
    a->expired_time = -1;
    a->lock_time = -1;
    a->backend_time = -1;
    a->backend_first_read = -1;
    a->backend_cache_time = -1;

    ap_ctx_set(r->ctx, "mod_accel", a);

    if (!(ab = ap_ctx_get(a->client_fb->ctx, "mod_accel_buff"))) {
        ab = ap_pcalloc(a->client_fb->pool, sizeof(accel_buff_rec));
        /* 14 accel log notes */
        ab->log_notes = ap_make_table(a->client_fb->pool, 14);

        ap_ctx_set(a->client_fb->ctx, "mod_accel_buff", ab);
    }

    ab->a = a;
    a->ab = ab;

    ap_register_cleanup(r->pool, (void *)a, accel_cleanup, ap_null_cleanup);

    return a;
}

void accel_init_dir_config(accel_rec *a, accel_dir_conf *conf)
{
    init(conf->unlink_no_cached, 0);
    init(conf->no_cache, 0);
    init(conf->ignore_no_cache, 0);
    init(conf->ignore_auth, 0);
    init(conf->revalidate_user, 0);
    init(conf->pass_cookie, 0);

    init(conf->cache_all_cookies, 0);
    if (conf->cache_cookie == UNSET)
        conf->cache_cookie = (conf->cache_all_cookies
                              || conf->cache_cookies->nelts) ? 1 : 0;

    init(conf->request_buff_size, REQUEST_BUFF_SIZE);
    init(conf->bk_rcv_buff_size, BK_RCV_BUFF_SIZE);
    init(conf->bk_rcv_sock_buff_size, 0);
    init(conf->timeout, a->r->server->timeout);
    init(conf->connect_timeout, conf->timeout);
    init(conf->retry_5XX, 1);
    init(conf->wait_before_body_read, 0);
    init(conf->send_size, SEND_SIZE);
    init(conf->cache_send_size, CACHE_SEND_SIZE);

    init(conf->ignore_expires, 0);

    init(conf->x_fwd_for, 0);
    init(conf->via, 0);
    init(conf->x_real_ip, 0);
    init(conf->x_host, 0);
    init(conf->x_url, 0);
    init(conf->pass_server, 0);

    init(conf->max_stale, 0);
    init(conf->busy_lock, 0);
    init(conf->expired_busy_lock, 0);
    init(conf->max_busy_lock, conf->timeout);
}

static void accel_cleanup(void *data)
{
    accel_buff_rec *ab;
    accel_rec *a = (accel_rec *)data;

    /* we remove tempfile in cleanup because ap_send_http_header() may be
       ap_hard_timeout()ed */

    if (a->tempfile && unlink(a->tempfile) == -1)
        ap_log_rerror(APLOG_MARK, APLOG_ERR, a->r,
                      "accel: error deleting temporary file %s", a->tempfile);

    /* clear busy lock */
    accel_clear_busy_lock(a);

    /* free malloc()ed memory */
    accel_cleanup_allocated(a->allocated);

    /* BUFF's CTX ab->a points to memory allocated from r->pool so we need
        to set it to NULL before r->pool would be freed */

    ab = ap_ctx_get(a->client_fb->ctx, "mod_accel_buff");
    ab->a = NULL;

    /* second SSI virtual include doesn't clean up r->ctx */
    if (a->r->main)
        ap_ctx_set(a->r->ctx, "mod_accel", NULL);
}


/* Config handlers */

static void *accel_create_server_config(pool *p, server_rec *s)
{
    int *level;
    accel_server_conf *conf = ap_pcalloc(p, sizeof(accel_server_conf));

    conf->pass = ap_make_array(p, 10, sizeof(accel_pass));
    conf->no_pass = ap_make_array(p, 10, sizeof(accel_no_pass));
    conf->reverse = ap_make_array(p, 10, sizeof(accel_reverse));

    conf->cache_root = NULL;

    conf->levels = ap_make_array(p, 4, sizeof(int));
    level = ap_push_array(conf->levels);
    *(level) = 1;
    level = ap_push_array(conf->levels);
    *(level) = 2;

    conf->auto_create_path = 1;

    conf->clean_at_time = UNSET;
    conf->clean_time = 3600;
    conf->clean_la = 24 * 3600;

    return (void *)conf;
}

static void *accel_merge_server_config(pool *p, void *base, void *new)
{
    accel_server_conf *merged = ap_pcalloc(p, sizeof(accel_server_conf));
    accel_server_conf *parent = (accel_server_conf *) base;
    accel_server_conf *child = (accel_server_conf *) new;

    merged->pass = ap_append_arrays(p, child->pass, parent->pass);
    merged->no_pass = ap_append_arrays(p, parent->no_pass, child->no_pass);
    merged->reverse = ap_append_arrays(p, parent->reverse, child->reverse);

    /* all these values has meaning in main server only */
    merged->cache_root = parent->cache_root ?
                         parent->cache_root : child->cache_root;
    merged->levels = parent->levels->nelts ? parent->levels : child->levels;
    merged->auto_create_path = parent->auto_create_path;

    merged->clean_at_time = parent->clean_at_time == UNSET ?
                            child->clean_at_time : parent->clean_at_time;
    merged->clean_time = parent->clean_time;
    merged->clean_la = parent->clean_la;

    return (void *)merged;
}

static void *accel_create_dir_config(pool *p, char *dir)
{
    accel_dir_conf *conf = ap_pcalloc(p, sizeof(accel_dir_conf));

    conf->unlink_no_cached = UNSET;
    conf->no_cache = UNSET;
    conf->ignore_no_cache = UNSET;
    conf->ignore_auth = UNSET;
    conf->revalidate_user = UNSET;
    conf->pass_cookie = UNSET;
    conf->cache_cookie = UNSET;
    conf->cache_all_cookies = UNSET;
    conf->cache_cookies = ap_make_array(p, 5, sizeof(accel_cookie));
    conf->no_cache_cookies = ap_make_array(p, 5, sizeof(accel_cookie));

    conf->request_buff_size = UNSET;
    conf->bk_rcv_buff_size = UNSET;
    conf->bk_rcv_sock_buff_size = UNSET;
    conf->timeout = UNSET;
    conf->connect_timeout = UNSET;
    conf->retry_5XX = UNSET;
    conf->wait_before_body_read = UNSET;
    conf->send_size = UNSET;
    conf->cache_send_size = UNSET;

    conf->cache_lm_factor = UNSET;
    conf->default_expire = UNSET;
    conf->ignore_expires = UNSET;
    conf->invl_len = UNSET;

    conf->x_fwd_for = UNSET;
    conf->via = UNSET;
    conf->x_real_ip = UNSET;
    conf->x_host = UNSET;
    conf->x_url = UNSET;
    conf->pass_server = UNSET;

    conf->max_stale = UNSET;
    conf->busy_lock = UNSET;
    conf->expired_busy_lock = UNSET;
    conf->max_busy_lock = UNSET;

    return (void *)conf;
}

static void *accel_merge_dir_config(pool *p, void *base, void *new)
{
    accel_dir_conf *merged = ap_pcalloc(p, sizeof(accel_dir_conf));
    accel_dir_conf *parent = (accel_dir_conf *) base;
    accel_dir_conf *child = (accel_dir_conf *) new;

    merge(merged, parent, child, unlink_no_cached);
    merge(merged, parent, child, no_cache);
    merge(merged, parent, child, ignore_no_cache);
    merge(merged, parent, child, ignore_auth);
    merge(merged, parent, child, revalidate_user);
    merge(merged, parent, child, pass_cookie);

    if (child->cache_cookie == UNSET
        && child->cache_all_cookies == UNSET
        && child->cache_cookies->nelts == 0
        && child->no_cache_cookies->nelts == 0)
    {
        merged->cache_cookie = parent->cache_cookie;
        merged->cache_all_cookies = parent->cache_all_cookies;
        merged->cache_cookies = parent->cache_cookies;
        merged->no_cache_cookies = parent->no_cache_cookies;

    } else {
        merged->cache_cookie = child->cache_cookie;
        merged->cache_all_cookies = child->cache_all_cookies;
        merged->cache_cookies = child->cache_cookies;
        merged->no_cache_cookies = child->no_cache_cookies;
    }

    merge(merged, parent, child, request_buff_size);
    merge(merged, parent, child, bk_rcv_buff_size);
    merge(merged, parent, child, bk_rcv_sock_buff_size);
    merge(merged, parent, child, timeout);
    merge(merged, parent, child, connect_timeout);
    merge(merged, parent, child, retry_5XX);
    merge(merged, parent, child, wait_before_body_read);
    merge(merged, parent, child, send_size);
    merge(merged, parent, child, cache_send_size);

    merge(merged, parent, child, cache_lm_factor);
    merge(merged, parent, child, default_expire);
    merge(merged, parent, child, ignore_expires);
    merged->invalidate = child->invl_len == UNSET ?
                         parent->invalidate : child->invalidate;
    merge(merged, parent, child, invl_len);

    merge(merged, parent, child, x_fwd_for);
    merge(merged, parent, child, via);
    merge(merged, parent, child, x_real_ip);
    merge(merged, parent, child, x_host);
    merge(merged, parent, child, x_url);
    merge(merged, parent, child, pass_server);

    merge(merged, parent, child, max_stale);
    merge(merged, parent, child, busy_lock);
    merge(merged, parent, child, expired_busy_lock);
    merged->max_busy_lock = child->max_busy_lock == UNSET ?
                            child->timeout : child->max_busy_lock;

    return (void *)merged;
}


static const char *add_pass(cmd_parms *cmd, void *dummy,
                            char *loc, char *url, char *flags)
{
    backend_rec     *b = NULL;

    accel_server_conf *conf = (accel_server_conf *)
               ap_get_module_config(cmd->server->module_config, &accel_module);

    const char *tag = NULL;

    accel_pass *pass = ap_push_array(conf->pass);
    pass->max_conn = 0;
    pass->max_wait = 0;
    pass->sum_part = ACCEL_SUM_NONE;
    pass->sum = 0;
    pass->loc_len = strlen(loc);
    pass->url_len = strlen(url);
    pass->loc = loc;
    pass->url = url;
    pass->addr.s_addr = INADDR_NONE;

    if (flags) {
        char *token;
        const char *ptr = flags;

        if (*ptr != '[' || flags[strlen(flags) - 1] != ']')
            return ap_psprintf(cmd->pool, "Invalid flag %s in AccelPass",
                                          flags);

        ptr++;

        while (*ptr && (token = ap_get_token(cmd->pool, &ptr, 0))) {
            int len = strlen(token);

            if (*ptr == ',' || *ptr == ';')         /* ap_get_token() bug */
                ptr++;

            if (len && token[len - 1] == ']')
                token[len - 1] = '\0';

            if (strncmp(token, "NR", 2) == 0) {
                pass->addr.s_addr = 0;
                continue;
            }

            if (strncmp(token, "PH", 2) == 0) {
                char *prefix, *port;
                if (!(prefix = strstr(url, "://")))
                    return "Invalid URL prefix";

                prefix += 3;

                pass->port = (port = strchr(prefix, ':')) ? atoi(port + 1) : 80;
                pass->preserve_host = 1;
                continue;
            }

            if (strncmp(token, "MC=", 3) == 0)
#ifdef EAPI_MM
                if ((pass->max_conn = atoi(&token[3])) > 0)
                    continue;
#else
                return "Flag MC can be used only "
                       "when EAPI compiled with MM library support";
#endif

            if (strncmp(token, "MW=", 3) == 0)
#ifdef EAPI_MM
                if ((pass->max_wait = atoi(&token[3])) > 0)
                    continue;
#else
                return "Flag MW can be used only "
                       "when EAPI compiled with MM library support";
#endif

            if (strncmp(token, "MP=", 3) == 0) {
#ifdef EAPI_MM
                if (token[3] == 'H') {
                    pass->sum_part = ACCEL_SUM_HOST;
                    continue;

                } else if (token[3] == 'P') {
                    pass->sum_part = ACCEL_SUM_PORT;
                    continue;

                } else if (token[3] == 'L') {
                    pass->sum_part = ACCEL_SUM_LOCATION;
                    continue;

                } else if (token[3] == 'T' && token[4] != '\0') {
                    pass->sum_part = ACCEL_SUM_TAG;
                    tag = &token[4];
                    continue;

                } else {
                    return ap_psprintf(cmd->pool,
                                       "Invalid value %s in AccelPass MP flag",
                                       token);
                }
#else
                return "Flag MP can be used only "
                       "when EAPI compiled with MM library support";
#endif
            }

            return ap_psprintf(cmd->pool, "Invalid flag %s in AccelPass",
                                          token);
        }
    }

    if (strncasecmp(url, "http://", 7) == 0) {
        if ((b = accel_parse_http_url(cmd->pool, url)) == NULL)
            return ap_psprintf(cmd->pool, "Invalid URL %s", url);

        if (strcmp(b->host, "_the_same_host_") == 0) {
            pass->port = b->port;
            pass->addr.s_addr = 0;
            pass->preserve_host = 1;
            pass->the_same_host = 1;
        }
    }

    if (b && pass->addr.s_addr == INADDR_NONE) {
        if ((pass->addr.s_addr = ap_inet_addr(b->host)) == INADDR_NONE) {
            struct hostent  *h;

            if ((h = ap_pgethostbyname(cmd->pool, b->host)) != NULL
                 && h->h_addr_list[0] != NULL)
            {
                pass->addr = *(struct in_addr *)(h->h_addr_list[0]);

            } else {
                return ap_psprintf(cmd->pool, "Can't resolve host %s", b->host);
            }
        }
    }

#ifdef EAPI_MM
    if (!pass->max_conn && pass->max_wait)
        return "Flag MW can be used with flag MC only";

    if (pass->max_conn && !pass->max_wait)
        pass->max_wait = pass->max_conn;

    if (!pass->max_conn && pass->sum_part != ACCEL_SUM_NONE)
        return "Flag MP can be used with flag MC only";

    if (pass->max_conn) {
        if (!pass->preserve_host && pass->sum_part == ACCEL_SUM_NONE)
            pass->sum_part = ACCEL_SUM_HOST;

        pass->sum = accel_sum(pass->sum_part, url, tag);
    }
#endif

    return NULL;
}

static const char *add_no_pass(cmd_parms *cmd, void *dummy, char *string)
{
    int   flag = REG_EXTENDED;
    char *regexp = NULL;

    accel_server_conf *conf = (accel_server_conf *)
               ap_get_module_config(cmd->server->module_config, &accel_module);

    accel_no_pass *no_pass = ap_push_array(conf->no_pass);

    if (string[0] == '~') {
        regexp = string + 1;

    } else {
        no_pass->len = strlen(string);
        no_pass->loc = string;
        return NULL;
    }

    if (regexp[0] == '*') {
        flag = REG_EXTENDED|REG_ICASE;
        regexp++;
    }

    if (regexp[0] == '\0')
        return "Invalid regexp in AccelNoPass";

    no_pass->regexp = ap_pregcomp(cmd->pool, regexp, flag);
    return NULL;
}

static const char *add_reverse(cmd_parms *cmd, void *dummy,
                               char *loc, char *url)
{
    accel_server_conf *conf = (accel_server_conf *)
               ap_get_module_config(cmd->server->module_config, &accel_module);

    accel_reverse *reverse = ap_push_array(conf->reverse);
    reverse->loc_len = strlen(loc);
    reverse->url_len = strlen(url);
    reverse->loc = loc;
    reverse->url = url;

    return NULL;
}

static const char *set_cache_root(cmd_parms *cmd, void *dummy, char *args)
{
    int  *level;
    char *value;
    const char *err;
    char *ptr = args;
    int   error = 0;
    int   first = 1;
    int   flag = 0;

    accel_server_conf *conf = (accel_server_conf *)
               ap_get_module_config(cmd->server->module_config, &accel_module);

    if ((err = ap_check_cmd_context(cmd, GLOBAL_ONLY)) != NULL)
        return err;

    if ((value = ap_getword_white_nc(cmd->pool, &ptr)) && *value)
        conf->cache_root = ap_server_root_relative(cmd->pool, value);
    else
        return "Invalid AccelCacheRoot";

    while (*ptr && (value = ap_getword_white_nc(cmd->pool, &ptr))) {

        if (strcmp(value, "noauto") == 0) {
            conf->auto_create_path = 0;
            flag = 1;
            continue;
        }

        if (flag)
            return "Invalid flag in AccelCacheRoot";

        if (conf->levels->nelts >= 4)
            return "AccelCacheRoot's level depth must be no more 3";

        if (first) {
            conf->levels->nelts = 0;
            first = 0;
        }

        level = ap_push_array(conf->levels);
        if (value[1] == '\0' && value[0] > '0' && value[0] < '3')
            *level = value[0] - '0';
        else
            error = 1;
    }

    if (error)
        return "Levels in AccelCacheRoot must has value of 1 or 2";

    return NULL;
}

static const char *set_clean(cmd_parms *cmd, void *dummy, const char *string)
{
    const char *err;

    accel_server_conf *conf = (accel_server_conf *)
               ap_get_module_config(cmd->server->module_config, &accel_module);

    if ((err = ap_check_cmd_context(cmd, GLOBAL_ONLY)) != NULL)
        return err;

    if (string[0] == '@')
        return accel_set_time_string(cmd->pool, cmd->cmd->name,
                                     ++string, &conf->clean_at_time);
    else
        return accel_set_time_string(cmd->pool, cmd->cmd->name,
                                     string, &conf->clean_time);
}

static const char *set_clean_last_accessed(cmd_parms *cmd, void *dummy,
                                           const char *string)
{
    const char *err;

    accel_server_conf *conf = (accel_server_conf *)
               ap_get_module_config(cmd->server->module_config, &accel_module);

    if ((err = ap_check_cmd_context(cmd, GLOBAL_ONLY)) != NULL)
        return err;

    return accel_set_time_string(cmd->pool, cmd->cmd->name,
                                 string, &conf->clean_la);
}

static const char *set_time_slot(cmd_parms *cmd,
                                  char *struct_ptr, char *string)
{
    int value;
    int offset = (int) (long) cmd->info;
    char *rv = accel_set_time_string(cmd->pool, cmd->cmd->name, string, &value);
    if (rv)
        return rv;

    *(int *) (struct_ptr + offset) = value;

    return NULL;
}

static const char *set_accel_timeout(cmd_parms *cmd, void *mconfig,
                               char *timeout, char *connect)
{
    accel_dir_conf *conf = (accel_dir_conf *)mconfig;

    char *rv = accel_set_time_string(cmd->pool, cmd->cmd->name,
                                     timeout, &conf->timeout);
    if (rv)
        return rv;

    if (connect)
        return accel_set_time_string(cmd->pool, cmd->cmd->name,
                                     connect, &conf->connect_timeout);

    conf->connect_timeout = conf->timeout;
    return NULL;
}

static const char *set_invalidate(cmd_parms *cmd, void *mconfig, char *string)
{
    accel_dir_conf *conf = (accel_dir_conf *)mconfig;

    if (strcasecmp(string, "off") == 0) {
        conf->invalidate = NULL;
        conf->invl_len = 0;
    } else {
        conf->invalidate = string;
        conf->invl_len = strlen(string);
    }

    return NULL;
}

static const char *cache_cookie(cmd_parms *cmd, void *mconfig, char *cookie)
{
    accel_dir_conf *conf = (accel_dir_conf *)mconfig;

    if (strcasecmp(cookie, "off") == 0) {
        conf->cache_cookie = 0;

    } else if (strcasecmp(cookie, "all") == 0) {
        conf->cache_all_cookies = 1;

    } else {
        accel_cookie *new;
        int i = 0;

        if (cookie[0] == '!') {
            new = ap_push_array(conf->no_cache_cookies);
            i = 1;

        } else {
            new = ap_push_array(conf->cache_cookies);
        }

        if (cookie[i] == '~') {
            int flag = REG_EXTENDED; 

            if (cookie[++i] == '*') {
                flag = REG_EXTENDED|REG_ICASE; 
                i++;
            }

            if (cookie[i] == '\0')
                return "Invalid regexp in AccelCacheCookie";

            new->regexp = ap_pregcomp(cmd->pool, &cookie[i], flag);

        } else {
            if (cookie[i] == '\0')
                return "Invalid string in AccelCacheCookie";

            new->name = &cookie[i];
        }
    }

    return NULL;
}

static const char *set_busy_lock(cmd_parms *cmd, void *mconfig,
                                 char *time, char *expired, char *max)
{
#ifdef EAPI_MM
    accel_dir_conf *conf = (accel_dir_conf *)mconfig;

    char *rv = accel_set_time_string(cmd->pool, cmd->cmd->name,
                                     time, &conf->busy_lock);
    if (rv)
        return rv;

    if (expired) {
        if (rv = accel_set_time_string(cmd->pool, cmd->cmd->name,
                                       expired, &conf->expired_busy_lock))
            return rv;

    } else {
        conf->expired_busy_lock = conf->busy_lock;
        return NULL;
    }

    if (max)
        return accel_set_time_string(cmd->pool, cmd->cmd->name,
                                     expired, &conf->max_busy_lock);
    return NULL;
#else
    return "AccelBusyLock can be used only "
           "when EAPI compiled with MM library support";
#endif
}



static const handler_rec accel_handlers[] =
{
    {"accel-handler", accel_handler},
#ifdef MOD_ACCEL_CACHEMGR
    {"accel-cachemgr", accel_cachemgr},
#endif
    {NULL}
};

static const command_rec accel_cmds[] =
{
    /* main and virtual server only */
    {"AccelPass", add_pass, NULL, RSRC_CONF, TAKE23,
     "set virtual path, URL and optional flags"},

    {"AccelNoPass", add_no_pass, NULL, RSRC_CONF, ITERATE,
     "set virtual path or regexp that should not be passed"},

    {"AccelReverse", add_reverse, NULL, RSRC_CONF, TAKE2,
     "set virtual path and URL to reverse proxiying"},


    /* main server only */
    {"AccelCacheRoot", set_cache_root, NULL, RSRC_CONF, RAW_ARGS,
     "set directory to store cache and temporary files, "
     "define up to 3 cache directory levels "
     "and set to create automatically directories in cache or not"},

    {"AccelClean", set_clean, NULL, RSRC_CONF, TAKE1,
     "set time to clean cache in time string"},

    {"AccelCleanLastAccessed", set_clean_last_accessed,
     NULL, RSRC_CONF, TAKE1,
     "set max last accessed time to clean cache in time string"},


    /* server and location */
    {"AccelNoCache", ap_set_flag_slot,
     (void *)XtOffsetOf(accel_dir_conf, no_cache),
     RSRC_CONF | ACCESS_CONF, FLAG,
     "disable caching"},

    {"AccelIgnoreNoCache", ap_set_flag_slot,
     (void *)XtOffsetOf(accel_dir_conf, ignore_no_cache),
     RSRC_CONF | ACCESS_CONF, FLAG,
     "set to ignore 'Pragma: no-cache' and so"},

    {"AccelIgnoreAuth", ap_set_flag_slot,
     (void *)XtOffsetOf(accel_dir_conf, ignore_auth),
     RSRC_CONF | ACCESS_CONF, FLAG,
     "set to ignore 'Authorization' header"},

    {"AccelRevalidateUser", ap_set_flag_slot,
     (void *)XtOffsetOf(accel_dir_conf, revalidate_user),
     RSRC_CONF | ACCESS_CONF, FLAG,
     "allow to revalidate user"},

    {"AccelPassCookie", ap_set_flag_slot,
     (void *)XtOffsetOf(accel_dir_conf, pass_cookie),
     RSRC_CONF | ACCESS_CONF, FLAG,
     "allow to pass Cookie to backend even request can be cachable"},

    {"AccelCacheCookie", cache_cookie, NULL,
     RSRC_CONF | ACCESS_CONF, ITERATE,
     "allow to cache cookie"},

    {"AccelRequestBuffSize", ap_set_kilobytes_slot,
     (void *)XtOffsetOf(accel_dir_conf, request_buff_size),
     RSRC_CONF | ACCESS_CONF, TAKE1,
     "set client request buffer size in Kbytes"},

    {"AccelBkRcvSockBuffSize", ap_set_kilobytes_slot,
     (void *)XtOffsetOf(accel_dir_conf, bk_rcv_sock_buff_size),
     RSRC_CONF | ACCESS_CONF, TAKE1,
     "set backend receive socket buffer size in Kbytes"},

    {"AccelBkRcvBuffSize", ap_set_kilobytes_slot,
     (void *)XtOffsetOf(accel_dir_conf, bk_rcv_buff_size),
     RSRC_CONF | ACCESS_CONF, TAKE1,
     "set backend receive buffer size in Kbytes"},

    {"AccelTimeout", set_accel_timeout, NULL, RSRC_CONF | ACCESS_CONF, TAKE12,
     "set backend timeout and connect timeout in time strings"},

    {"AccelRetry5XX", ap_set_flag_slot,
     (void *)XtOffsetOf(accel_dir_conf, retry_5XX),
     RSRC_CONF | ACCESS_CONF, FLAG,
     "retry 5XX result if there are several backends"},

    {"AccelWaitBeforeBodyRead", ap_set_integer_slot,
     (void *)XtOffsetOf(accel_dir_conf, wait_before_body_read),
     RSRC_CONF | ACCESS_CONF, TAKE1,
     "set time to wait before read body from backend in milliseconds"},

    {"AccelSendSize", ap_set_kilobytes_slot,
     (void *)XtOffsetOf(accel_dir_conf, send_size),
     RSRC_CONF | ACCESS_CONF, TAKE1,
     "set client send buffer size in Kbytes"},

    {"AccelCacheSendSize", ap_set_kilobytes_slot,
     (void *)XtOffsetOf(accel_dir_conf, cache_send_size),
     RSRC_CONF | ACCESS_CONF, TAKE1,
     "set client send buffer size in Kbytes when response is cached"},

    {"AccelLastModifiedFactor", ap_set_integer_slot,
     (void *)XtOffsetOf(accel_dir_conf, cache_lm_factor),
     RSRC_CONF | ACCESS_CONF, TAKE1,
     "set LastModified expire factor in percent"},

    {"AccelDefaultExpire", set_time_slot,
     (void *)XtOffsetOf(accel_dir_conf, default_expire),
     RSRC_CONF | ACCESS_CONF, TAKE1,
     "set default expire in time string"},

    {"AccelIgnoreExpires", ap_set_flag_slot,
     (void *)XtOffsetOf(accel_dir_conf, ignore_expires),
     RSRC_CONF | ACCESS_CONF, FLAG,
     "set to ignore Expires header from backend"},

    {"AccelInvalidate", set_invalidate, NULL,
     RSRC_CONF | ACCESS_CONF, TAKE1,
     "set invalidate string to refresh cache file"},

    {"AccelUnlinkNoCached", ap_set_flag_slot,
     (void *)XtOffsetOf(accel_dir_conf, unlink_no_cached),
     RSRC_CONF | ACCESS_CONF, FLAG,
     "set to unlink no cached respose always"},

    {"AccelAddXForwardedFor", ap_set_flag_slot,
     (void *)XtOffsetOf(accel_dir_conf, x_fwd_for),
     RSRC_CONF | ACCESS_CONF, FLAG,
     "add client's IP address to X-Forwarded-For header"},

    {"AccelAddVia", ap_set_flag_slot,
     (void *)XtOffsetOf(accel_dir_conf, via),
     RSRC_CONF | ACCESS_CONF, FLAG,
     "add Via header"},

    {"AccelSetXRealIP", ap_set_flag_slot,
     (void *)XtOffsetOf(accel_dir_conf, x_real_ip),
     RSRC_CONF | ACCESS_CONF, FLAG,
     "set X-Real-IP header to client's IP address"},

    {"AccelSetXHost", ap_set_flag_slot,
     (void *)XtOffsetOf(accel_dir_conf, x_host),
     RSRC_CONF | ACCESS_CONF, FLAG,
     "set X-Host header to client's Host header"},

    {"AccelSetXURL", ap_set_flag_slot,
     (void *)XtOffsetOf(accel_dir_conf, x_url),
     RSRC_CONF | ACCESS_CONF, FLAG,
     "set X-URL header to client's URL"},

    {"AccelSetXURI", ap_set_flag_slot,
     (void *)XtOffsetOf(accel_dir_conf, x_uri),
     RSRC_CONF | ACCESS_CONF, FLAG,
     "set X-URI header to client's URI"},

    {"AccelPassServer", ap_set_flag_slot,
     (void *)XtOffsetOf(accel_dir_conf, pass_server),
     RSRC_CONF | ACCESS_CONF, FLAG,
     "pass backend Server header"},

    {"AccelMaxStale", set_time_slot,
     (void *)XtOffsetOf(accel_dir_conf, max_stale),
     RSRC_CONF | ACCESS_CONF, TAKE1,
     "set max stale response in time string"},

    {"AccelBusyLock", set_busy_lock, NULL,
     RSRC_CONF | ACCESS_CONF, TAKE123,
     "set concurrent cache update busy lock time in time string"},

    {NULL}
};

module accel_module =
{
    STANDARD_MODULE_STUFF,
    accel_init,                       /* module initializer */
    accel_create_dir_config,          /* per-directory config creator */
    accel_merge_dir_config,           /* dir config merger */
    accel_create_server_config,       /* server config creator */
    accel_merge_server_config,        /* server config merger */
    accel_cmds,                       /* command table */
    accel_handlers,                   /* [9] list of handlers */
    accel_translate,                  /* [2] filename-to-URI translation */
    NULL,                             /* [5] check/validate user_id */
    NULL,                             /* [6] check user_id is valid *here* */
    NULL,                             /* [4] check access by host address */
    NULL,                             /* [7] MIME type checker/setter */
    NULL,                             /* [8] fixups */
    NULL,                             /* [10] logger */
    NULL,                             /* [3] header parser */
    NULL,                             /* process initializer */
    NULL,                             /* process exit/cleanup */
    accel_postread,                   /* [1] post read_request handling */

    /* EAPI */
    accel_add_module,                 /* after modules was added to core */
    accel_remove_module               /* before module is removed from core */
};
